/*
 * Kernel Debugger Architecture Dependent Stack Traceback
 *
 * Copyright (C) 1999-2002 Silicon Graphics, Inc.  All Rights Reserved
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it would be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * Further, this software is distributed without any warranty that it is
 * free of the rightful claim of any third person regarding infringement
 * or the like.  Any license provided herein, whether implied or
 * otherwise, applies only to this software file.  Patent licenses, if
 * any, provided herein do not apply to combinations of this program with
 * other software, or any other product whatsoever.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston MA 02111-1307, USA.
 *
 * Contact information:  Silicon Graphics, Inc., 1600 Amphitheatre Pkwy,
 * Mountain View, CA  94043, or:
 *
 * http://www.sgi.com
 *
 * For further information regarding this notice, see:
 *
 * http://oss.sgi.com/projects/GenInfo/NoticeExplan
 */

#include <linux/config.h>
#include <linux/ctype.h>
#include <linux/string.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#ifdef CONFIG_SMP
#include <linux/sched_runqueue.h>
#endif
#include <linux/kallsyms.h>
#include <linux/kdb.h>
#include <linux/kdbprivate.h>
#include <asm/system.h>

/*
 * bt_print_one
 *
 *	Print one back trace entry.
 *
 * Inputs:
 *	ip	Current program counter.
 *	symtab	Information about symbol that ip falls within.
 *	ar	Activation record for this frame.
 *	argcount Maximum number of arguments to print.
 * Outputs:
 *	None.
 * Returns:
 *	None.
 * Locking:
 *	None.
 * Remarks:
 *	None.
 */

static void
bt_print_one(kdb_machreg_t ip, const kdb_ar_t *ar,
	     const kdb_symtab_t *symtab, int argcount,
	     struct unw_frame_info *info /* FIXME: should be part of ar for ia64 */)
{
	int	btsymarg = 0;		/* Convert arguments to symbols */
	int	btsp = 0;		/* Print stack and backing store pointers */
	int	nosect = 0;		/* Suppress section data */
	kdb_machreg_t sp, bsp, cfm;	/* FIXME: should be part of ar for ia64 */

	kdbgetintenv("BTSYMARG", &btsymarg);
	kdbgetintenv("BTSP", &btsp);
	kdbgetintenv("NOSECT", &nosect);

	unw_get_sp(info, &sp);		/* FIXME: should be part of ar for ia64 */
	unw_get_bsp(info, &bsp);	/* FIXME: should be part of ar for ia64 */
	unw_get_cfm(info, &cfm);	/* FIXME: info/cfm should be part of ar for ia64 */
	kdb_symbol_print(ip, symtab, KDB_SP_VALUE|KDB_SP_NEWLINE);
	/* FIXME: number of args should be set in prologue code */
	((kdb_ar_t *)ar)->args = (cfm >> 7) & 0x7f;	/* sol */
	if (!ar->args)
		((kdb_ar_t *)ar)->args = cfm & 0x7f;	/* no in/local, use sof instead */
	if (argcount && ar->args) {
		int i, argc = ar->args;

		kdb_printf("        args (");
		if (argc > argcount)
			argc = argcount;

		for(i = 0; i < argc; i++){
			/* FIXME: prologue code should extract arguments */
			kdb_machreg_t arg;
			char nat;
			if (unw_access_gr(info, i+32, &arg, &nat, 0))
				arg = 0;

			if (i)
				kdb_printf(", ");
			kdb_printf("0x%lx", arg);
		}
		kdb_printf(")\n");
		if (btsymarg) {
			kdb_symtab_t	arg_symtab;
			kdb_machreg_t	arg;
			for(i = 0; i < argc; i++){
				/* FIXME: prologue code should extract arguments */
				char nat;
				if (unw_access_gr(info, i+32, &arg, &nat, 0))
					arg = 0;
				if (kdbnearsym(arg, &arg_symtab)) {
					kdb_printf("        arg %d ", i);
					kdb_symbol_print(arg, &arg_symtab, KDB_SP_DEFAULT|KDB_SP_NEWLINE);
				}
			}
		}
	}
	if (symtab->sym_name) {
		if (!nosect) {
			kdb_printf("        %s %s 0x%lx 0x%lx 0x%lx\n",
				symtab->mod_name,
				symtab->sec_name,
				symtab->sec_start,
				symtab->sym_start,
				symtab->sym_end);
		}
	}
	if (btsp)
		kdb_printf("        sp 0x%016lx bsp 0x%016lx cfm 0x%016lx\n", sp, bsp, cfm);
}

/*
 * kdba_bt_stack
 *
 *	This function implements the 'bt' command.  Print a stack
 *	traceback.
 *
 *	bt                           (stack address is not supported on IA64)
 *	btp <pid>		     (Kernel stack for <pid>)
 *
 * Inputs:
 *	regs	registers at time kdb was entered.
 *	addr	Pointer to Address provided to 'bt' command, if any.
 *	argcount
 *	p	Pointer to task for 'btp' command.
 * Outputs:
 *	None.
 * Returns:
 *	zero for success, a kdb diagnostic if error
 * Locking:
 *	none.
 * Remarks:
 *	mds comes in handy when examining the stack to do a manual
 *	traceback.
 */

int
kdba_bt_stack(struct pt_regs *regs, kdb_machreg_t *addr, int argcount,
	      struct task_struct *p)
{
	kdb_symtab_t symtab;
	kdb_ar_t	ar;
	struct unw_frame_info info;	/* FIXME: should be part of ar */
	struct switch_stack *sw;	/* FIXME: should be part of ar */
#ifdef CONFIG_SMP
	runqueue_t *rq;
#endif

	if (addr) {
		kdb_printf("bt <address> is unsupported for IA-64\n");
		return 0;
	}

	/* FIXME: All the arch specific code should be in activation records, not here */
	memset(&ar, 0, sizeof(ar));

	/*
	 * Upon entering kdb_main_loop, the stack frame looks like this:
	 *
	 *	+---------------------+
	 *	|   struct pt_regs    |
	 *	+---------------------+
	 *	|		      |
	 *	|   kernel stack      |
	 *	|		      |
	 *	+=====================+ <--- top of stack upon entering kdb
	 *	|   struct pt_regs    |
	 *	+---------------------+
	 *	|		      |
	 *	|   kdb stack         |
	 *	|		      |
	 *	+---------------------+
	 *	| struct switch_stack |
	 *	+=====================+ <--- kdb_sw[cpu] from do_kdba_main_loop
	 *
	 * When looking at another process, we do not have the address of the
	 * current pt_regs, it is NULL.
	 */

	sw = NULL;
#ifdef CONFIG_SMP
	rq = task_rq(p);
#endif
	if (p == current) {
		sw = kdb_sw[smp_processor_id()];
	}
#ifdef CONFIG_SMP
	else if (task_running(rq, p)) {
		sw = kdb_sw[p->cpu];
	}
#endif
	else {
		/* Not running, assume blocked */
	        sw = (struct switch_stack *) (p->thread.ksp + 16);
	}
	if (!sw) {
		kdb_printf("Process does not have a switch_stack, cannot backtrace\n");
		return 0;
	}

	unw_init_frame_info(&info, p, sw);	/* FIXME: should be using activation records */

	/* If we have the address of pt_regs, suppress backtrace on the frames below
	 * pt_regs.  No point in displaying kdb itself.
	 */
	if (regs) {
		kdb_machreg_t sp;		/* FIXME: should be part of ar for ia64 */
		if (user_mode(regs)) {
			kdb_printf("Process was interrupted in user mode, no backtrace available\n");
			return 0;
		}
		do {
			unw_get_sp(&info, &sp);
			if (sp >= (kdb_machreg_t)regs)
				break;
		} while (unw_unwind(&info) >= 0);	/* FIXME: should be using activation records */
	}

	do {
		kdb_machreg_t ip;

		unw_get_ip(&info, &ip);	/* FIXME: should be using activation records */
		if (ip == 0)
			break;

		kdbnearsym(ip, &symtab);
		if (!symtab.sym_name) {
			kdb_printf("0x%0*lx - No name.  May be an area that has no unwind data\n",
				(int)(2*sizeof(ip)), ip);
			return 0;
		}
		bt_print_one(ip, &ar, &symtab, argcount, &info);
	} while (unw_unwind(&info) >= 0);	/* FIXME: should be using activation records */

	return 0;
}

int
kdba_bt_process(struct task_struct *p, int argcount)
{
	return kdba_bt_stack(NULL, NULL, argcount, p);
}
