/* 
 * Copyright (C) 2002 Jeff Dike (jdike@karaya.com)
 * Licensed under the GPL
 */

#include <linux/fs.h>
#include <linux/module.h>
#include <linux/slab.h>
#include "os.h"

struct hppfs_private {
	int host_file;
	unsigned int host_sock:1;
	unsigned int host_filter:1;
	unsigned int sent_data:1;
	struct file proc_file;
};

#define HPPFS_SUPER_MAGIC 0xb00000ee

static struct inode *get_inode(struct super_block *sb, struct dentry *dentry,
			       int *error);

struct dentry_operations hppfs_dentry_ops = {
};

struct dentry *hppfs_lookup(struct inode *ino, struct dentry *dentry)
{
	struct dentry *proc_dentry;
	struct inode *inode;
	int err;

	proc_dentry = lookup_hash(&dentry->d_name, ino->u.hppfs_i.proc_dentry);
	if(IS_ERR(proc_dentry))
		return(proc_dentry);

	inode = get_inode(ino->i_sb, proc_dentry, &err);
	if(err != 0) 
		return(ERR_PTR(err));

	d_add(dentry, inode);
	dentry->d_op = &hppfs_dentry_ops;
	return(NULL);
}

static struct inode_operations hppfs_file_iops = {
};

static struct inode_operations hppfs_dir_iops = {
	lookup:		hppfs_lookup,
};

static ssize_t read_proc(struct file *file, struct file *hppfs_file, char *buf,
			 ssize_t count, loff_t *ppos)
{
	ssize_t (*read)(struct file *, char *, size_t, loff_t *);
	ssize_t n;

	read = file->f_dentry->d_inode->i_fop->read;
		
	file->f_pos = hppfs_file->f_pos;
	n = (*read)(file, buf, count, &file->f_pos);
	hppfs_file->f_pos = file->f_pos;
	if(ppos) *ppos = file->f_pos;
	return(n);
}

static ssize_t hppfs_read(struct file *file, char *buf, size_t count, 
			  loff_t *ppos)
{
	struct hppfs_private *data = file->private_data;
	int n, fd;

	fd = data->host_file;
	if(fd > 0){
		if(!data->host_sock){
			n = os_seek_file(fd, *ppos);
			if(n < 0)
				return(n);
		}
		else if(data->host_filter && !data->sent_data){
			while((n = read_proc(&data->proc_file, file, buf, 
					     count, NULL)) > 0)
				os_write_file(fd, buf, n);
			os_shutdown_socket(fd, 0, 1);
			data->sent_data = 1;
		}

		n = os_read_file(fd, buf, count);
		if(n < 0)
			return(n);

		*ppos += n;
	}
	else n = read_proc(&data->proc_file, file, buf, count, ppos);

	return(n);
}

static char *dentry_name(struct dentry *dentry, int extra)
{
	struct dentry *parent;
	char *root, *name;
	int len;

	len = 0;
	parent = dentry;
	while(parent->d_parent != parent){
		len += parent->d_name.len + 1;
		parent = parent->d_parent;
	}
	
	root = "proc";
	len += strlen(root);
	name = kmalloc(len + extra + 1, GFP_KERNEL);
	if(name == NULL) return(NULL);

	name[len] = '\0';
	parent = dentry;
	while(parent->d_parent != parent){
		len -= parent->d_name.len + 1;
		name[len] = '/';
		strncpy(&name[len + 1], parent->d_name.name, 
			parent->d_name.len);
		parent = parent->d_parent;
	}
	strncpy(name, root, strlen(root));
	return(name);
}

static int open_host_sock(char *host_file, int *filter_out)
{
	char *end;
	int fd;

	end = &host_file[strlen(host_file)];
	strcpy(end, "/rw");
	*filter_out = 1;
	fd = os_connect_socket(host_file);
	if(fd > 0)
		return(fd);

	strcpy(end, "/r");
	*filter_out = 0;
	fd = os_connect_socket(host_file);
	return(fd);
}

static int hppfs_open(struct inode *inode, struct file *file)
{
	struct hppfs_private *data;
	struct dentry *proc_dentry;
	char *host_file;
	int err, fd, type, filter;

	err = -ENOMEM;
	data = kmalloc(sizeof(*data), GFP_KERNEL);
	if(data == NULL)
		goto out;

	data->host_file = -1;
	host_file = dentry_name(file->f_dentry, strlen("/rw"));
	if(host_file == NULL)
		goto out_free2;

	type = os_file_type(host_file);
	if(type == OS_TYPE_FILE){
		fd = os_open_file(host_file, of_read(OPENFLAGS()), 0);
		if(fd >= 0) 
			data->host_file = fd;
		data->host_sock = 0;
	}
	else if(type == OS_TYPE_DIR){
		fd = open_host_sock(host_file, &filter);
		if(fd > 0){
			data->host_file = fd;
			data->host_sock = 1;
			data->sent_data = 0;
			data->host_filter = filter;
		}
	}
	kfree(host_file);

	proc_dentry = inode->u.hppfs_i.proc_dentry;
	err = init_private_file(&data->proc_file, proc_dentry, file->f_mode);
	if(err)
		goto out_free1;

	file->f_pos = data->proc_file.f_pos;
	file->private_data = data;
	return(0);

 out_free1:
	kfree(host_file);
 out_free2:
	kfree(data);
 out:
	return(err);
}

static int hppfs_dir_open(struct inode *inode, struct file *file)
{
	struct hppfs_private *data;
	struct dentry *proc_dentry;
	int err;

	err = -ENOMEM;
	data = kmalloc(sizeof(*data), GFP_KERNEL);
	if(data == NULL)
		goto out;
	data->host_file = -1;
	proc_dentry = inode->u.hppfs_i.proc_dentry;
	err = init_private_file(&data->proc_file, proc_dentry, file->f_mode);
	if(err)
		goto out_free;

	file->f_pos = data->proc_file.f_pos;
	file->private_data = data;
	return(0);

 out_free:
	kfree(data);
 out:
	return(err);
}

static int hppfs_readdir(struct file *file, void *ent, filldir_t filldir)
{
	struct hppfs_private *data = file->private_data;
	struct file *proc_file = &data->proc_file;
	int (*readdir)(struct file *, void *, filldir_t);
	int err;

	readdir = proc_file->f_dentry->d_inode->i_fop->readdir;

	proc_file->f_pos = file->f_pos;
	err = (*readdir)(proc_file, ent, filldir);
	file->f_pos = proc_file->f_pos;

	return(err);
}

static int hppfs_dir_release(struct inode *ino, struct file *file)
{
	return(0);
}

static int hppfs_fsync(struct file *file, struct dentry *dentry, int datasync)
{
	return(0);
}

static struct file_operations hppfs_file_fops = {
	owner:		NULL,
	read:		hppfs_read,
	open:		hppfs_open,
};

static struct file_operations hppfs_dir_fops = {
	owner:		NULL,
	readdir:	hppfs_readdir,
	open:		hppfs_dir_open,
	release:	hppfs_dir_release,
	fsync:		hppfs_fsync,
};

static int hppfs_statfs(struct super_block *sb, struct statfs *sf)
{
	sf->f_blocks = 0;
	sf->f_bfree = 0;
	sf->f_bavail = 0;
	sf->f_files = 0;
	sf->f_ffree = 0;
	sf->f_type = HPPFS_SUPER_MAGIC;
	return(0);
}

static struct super_operations hppfs_sbops = { 
	put_inode:	force_delete,
	delete_inode:	NULL,
	statfs:		hppfs_statfs,
};

static int hppfs_readlink(struct dentry *dentry, char *buffer, int buflen)
{
	struct file proc_file;
	struct dentry *proc_dentry;
	int (*readlink)(struct dentry *, char *, int);
	int err, n;

	proc_dentry = dentry->d_inode->u.hppfs_i.proc_dentry;
	err = init_private_file(&proc_file, proc_dentry, FMODE_READ);
	if(err) 
		return(err);

	readlink = proc_dentry->d_inode->i_op->readlink;
	n = (*readlink)(proc_dentry, buffer, buflen);

	if(proc_file.f_op->release)
		(*proc_file.f_op->release)(proc_dentry->d_inode, &proc_file);
	
	return(n);
}

static struct inode_operations hppfs_link_iops = {
	readlink:	hppfs_readlink,
};

static void read_inode(struct inode *ino)
{
	struct inode *proc_ino;

	proc_ino = ino->u.hppfs_i.proc_dentry->d_inode;
	ino->i_uid = proc_ino->i_uid;
	ino->i_gid = proc_ino->i_gid;
	ino->i_atime = proc_ino->i_atime;
	ino->i_mtime = proc_ino->i_mtime;
	ino->i_ctime = proc_ino->i_ctime;
	ino->i_ino = proc_ino->i_ino;
	ino->i_dev = proc_ino->i_dev;
	ino->i_mode = proc_ino->i_mode;
	ino->i_nlink = proc_ino->i_nlink;
	ino->i_size = proc_ino->i_size;
	ino->i_blksize = proc_ino->i_blksize;
	ino->i_blocks = proc_ino->i_blocks;
}

static struct inode *get_inode(struct super_block *sb, struct dentry *dentry,
			       int *error)
{
	struct inode *inode;
	int err = -ENOMEM;

	inode = get_empty_inode();
	if(inode == NULL) 
		goto out;

	insert_inode_hash(inode);
	if(S_ISDIR(dentry->d_inode->i_mode)){
		inode->i_op = &hppfs_dir_iops;
		inode->i_fop = &hppfs_dir_fops;
	}
	else if(S_ISLNK(dentry->d_inode->i_mode)){
		inode->i_op = &hppfs_link_iops;
		inode->i_fop = &hppfs_file_fops;
	}
	else {
		inode->i_op = &hppfs_file_iops;
		inode->i_fop = &hppfs_file_fops;
	}

	inode->i_sb = sb;
	inode->u.hppfs_i.proc_dentry = dentry;

	read_inode(inode);
	err = 0;

	if(error) *error = err;
	return(inode);
 out:
	if(error) *error = err;
	return(NULL);
}

static struct super_block *hppfs_read_super(struct super_block *sb, void *d, 
					    int silent)
{
	struct inode *root_inode;
	struct file_system_type *procfs;
	struct super_block *proc_sb;

	procfs = get_fs_type("proc");
	if(procfs == NULL) 
		goto out;

	if(list_empty(&procfs->fs_supers))
		goto out;

	proc_sb = list_entry(procfs->fs_supers.next, struct super_block,
			     s_instances);
	
	sb->s_blocksize = 1024;
	sb->s_blocksize_bits = 10;
	sb->s_magic = HPPFS_SUPER_MAGIC;
	sb->s_op = &hppfs_sbops;

	dget(proc_sb->s_root);
	root_inode = get_inode(sb, proc_sb->s_root, NULL);
	if(root_inode == NULL)
		goto out_dput;

	sb->s_root = d_alloc_root(root_inode);
	if(sb->s_root == NULL)
		goto out_put;

	return(sb);

 out_put:
	iput(root_inode);
 out_dput:
	dput(proc_sb->s_root);
 out:
	return(NULL);
}

DECLARE_FSTYPE(hppfs_type, "hppfs", hppfs_read_super, 0);

static int __init init_hppfs(void)
{
	return(register_filesystem(&hppfs_type));
}

static void __exit exit_hppfs(void)
{
	unregister_filesystem(&hppfs_type);
}

module_init(init_hppfs)
module_exit(exit_hppfs)

/*
 * Overrides for Emacs so that we follow Linus's tabbing style.
 * Emacs will notice this stuff at the end of the file and automatically
 * adjust the settings for this buffer only.  This must remain at the end
 * of the file.
 * ---------------------------------------------------------------------------
 * Local variables:
 * c-file-style: "linux"
 * End:
 */
