/*
 *  blacklist.c
 *
 *  Check to see if the given machine has a known bad ACPI BIOS
 *
 *  Copyright (C) 2002 Andy Grover <andrew.grover@intel.com>
 *
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
 *
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 */


#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/acpi.h>
#include "acpi_bus.h"

#define	PREFIX			"ACPI: "

enum acpi_blacklist_predicates
{
        all_versions,
        less_than_or_equal,
        equal,
        greater_than_or_equal,
};

struct acpi_blacklist_item
{
        char            oem_id[7];
        char            oem_table_id[9];
        u32             oem_revision;
        acpi_table_type table;
        enum acpi_blacklist_predicates oem_revision_predicate;
        char            *reason;
        u32             is_critical_error;
};

/*
 * POLICY: If *anything* doesn't work, put it on the blacklist.
 *	   If they are critical errors, mark it critical, and abort driver load.

EXTENDED POLICY: 

the UL kernel by default uses ACPI with ACPI PCI routing. 

When IO devices (e.g. SCSI controller)  do not work likely something is wrong
with the PCI routing. Try it with pci=noacpi then. If that helps runs
dmidecode and add an "disable_acpi_pci" entry to the blacklist in 
arch/i386/kernel/dmi_scan.c

When the box crashes very early try to boot with acpi=oldboot.
If that helps run dmidecode and add an "force_acpi_oldboot" entry to arch/i386/kernel/
dmi_scan.c

When this still doesn't help try it with acpi=off
If that helps run dmidecode and add an "acpi_disable" entry to arch/i386/kernel/dmi_scan.c

Any blacklists can be overwritten with acpi=force

Alternatively to running dmidecode you can also run acpidmp and add an
blacklist entry to drivers/acpi/blacklist.c for broken ACPI tables (e.g. usually the 
MSDT for PCI routing problems). Please only do that when the OEM entry/ID looks
unique - e.g. is a specific version and model number - not for generic entries
like INTEL AWRDPCI. The choices there are ACPI_OLDBOOT (acpi=oldboot), ACPI_BROKEN 
(acpi=off), ACPI_FORCE_PCI (obsolete, now default), ACPI_NO_PCI (pci=noacpi)

When acpi=oldboot is active the rest of ACPI is disabled.

Any blacklist entries can be overwritten to use full ACPI with acpi=force.
When there is an  ACPI_NO_PCI entry  or disable_acpi_pci entry you can overwrite
it with pci=acpi.

-AK

 */

static struct acpi_blacklist_item acpi_blacklist[] __initdata =
{
	/* Portege 7020, BIOS 8.10 */
	{"TOSHIB", "7020CT  ", 0x19991112, ACPI_DSDT, all_versions, "Implicit Return", 0},
	/* Portege 4030 */
	{"TOSHIB", "4030    ", 0x19991112, ACPI_DSDT, all_versions, "Implicit Return", 0},
	/* Portege 310/320, BIOS 7.1 */
	{"TOSHIB", "310     ", 0x19990511, ACPI_DSDT, all_versions, "Implicit Return", 0},
	/* Seattle 2, old bios rev. */
	{"INTEL ", "440BX   ", 0x00001000, ACPI_DSDT, less_than_or_equal, "Field beyond end of region", 0},
	/* ASUS K7M */
	{"ASUS  ", "K7M     ", 0x00001000, ACPI_DSDT, less_than_or_equal, "Field beyond end of region", 0},
	/* Intel 810 Motherboard? */
	{"MNTRAL", "MO81010A", 0x00000012, ACPI_DSDT, less_than_or_equal, "Field beyond end of region", 0},
	/* Compaq Presario 711FR */
	{"COMAPQ", "EAGLES", 0x06040000, ACPI_DSDT, less_than_or_equal, "SCI issues (C2 disabled)", 0},
	/* Compaq Presario 1700 */
	{"PTLTD ", "  DSDT  ", 0x06040000, ACPI_DSDT, less_than_or_equal, "Multiple problems", 1},
	/* Sony FX120, FX140, FX150? */
	{"SONY  ", "U0      ", 0x20010313, ACPI_DSDT, less_than_or_equal, "ACPI driver problem", 1},
	/* Compaq Presario 800, Insyde BIOS */
	{"INT440", "SYSFexxx", 0x00001001, ACPI_DSDT, less_than_or_equal, "Does not use _REG to protect EC OpRegions", 1},
	/* IBM 600E - _ADR should return 7, but it returns 1 */
	{"IBM   ", "TP600E  ", 0x00000105, ACPI_DSDT, less_than_or_equal, "Incorrect _ADR", 1},

	/* AK: should check if they work with ACPI_OLDBOOT or ACPI_PCI_FORCE */

	{"VIA694", "AWRDACPI", 0, ACPI_DSDT, all_versions, "Bogus table", 1},
	{"ASUS\0\0", "P2B-S   ", 0, ACPI_DSDT, all_versions, "Bogus PCI routing", ACPI_BROKEN},

	{"VIA694", "AWRDACPI", 0,          ACPI_DSDT, all_versions, "Bogus table", ACPI_BROKEN},
	{"ASUS",   "P2B-S   ", 0,          ACPI_DSDT, all_versions, "Bogus PCI routing", ACPI_BROKEN},
	{"IBM   ", "TP-16   ", 0,          ACPI_DSDT, all_versions, "Bogus PCI routing", ACPI_BROKEN},
	{"IBM   ", "TP-T21  ", 0,          ACPI_DSDT, all_versions, "Bogus PCI routing", ACPI_BROKEN},
	{"SHARP ", "QS1     ", 0,          ACPI_DSDT, all_versions, "Bogus table", ACPI_BROKEN},
	{"Intel",  "Trajan",   0,          ACPI_DSDT, all_versions, "Bogus table", ACPI_BROKEN},

	/* keyed on the FADT table for now. Hopefully the magic number 
	   won't change. */
	{"UNISYS", "440BX309",   0,	   ACPI_FADT, all_versions, 
	 "es7000 needs oldboot", ACPI_ES7000 }, 
	
	{"IBM   ", "SERVIGIL",   0,	   ACPI_DSDT, all_versions, 
	 "x440 needs oldboot", ACPI_OLDBOOT }, 
	
	{""}
};


enum acpi_black __init
acpi_blacklisted(void)
{
	int i = 0;
	int blacklisted = 0;
	struct acpi_table_header *table_header;

	while (acpi_blacklist[i].oem_id[0] != '\0')
	{
		if (acpi_get_table_header_early(acpi_blacklist[i].table, &table_header)) {
			i++;
			continue;
		}

		if (strncmp(acpi_blacklist[i].oem_id, table_header->oem_id, 6)) {
			i++;
			continue;
		}

		if (strncmp(acpi_blacklist[i].oem_table_id, table_header->oem_table_id, 8)) {
			i++;
			continue;
		}

		if ((acpi_blacklist[i].oem_revision_predicate == all_versions)
		    || (acpi_blacklist[i].oem_revision_predicate == less_than_or_equal
		        && table_header->oem_revision <= acpi_blacklist[i].oem_revision)
		    || (acpi_blacklist[i].oem_revision_predicate == greater_than_or_equal
		        && table_header->oem_revision >= acpi_blacklist[i].oem_revision)
		    || (acpi_blacklist[i].oem_revision_predicate == equal
		        && table_header->oem_revision == acpi_blacklist[i].oem_revision)) {

			printk(KERN_ERR PREFIX "Vendor \"%6.6s\" System \"%8.8s\" "
				"Revision 0x%x has a known ACPI BIOS problem.\n",
				acpi_blacklist[i].oem_id,
				acpi_blacklist[i].oem_table_id,
				acpi_blacklist[i].oem_revision);

			printk(KERN_ERR PREFIX "Reason: %s. This is a %s error\n",
				acpi_blacklist[i].reason,
				(acpi_blacklist[i].is_critical_error ? "non-recoverable" : "recoverable"));

			blacklisted = acpi_blacklist[i].is_critical_error;
			break;
		}
		else {
			i++;
		}
	}

	return blacklisted;
}

