/*
 * IBM Dynamic Probes
 * Copyright (c) International Business Machines Corp., 2000
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or 
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  
 */
#include <linux/dprobes.h>
#include <linux/hook.h>

void dprobes_hooks_code_start(void) { }

USE_HOOK(SYS_INIT_MODULE);
USE_HOOK(SYS_INIT_MODULE_ERROR);
USE_HOOK(FREE_MODULE);

static int insmod_exit(hook_rec_t * hookrec, struct module * mod)
{
	dp_insmod(mod);
	return HOOK_CONTINUE;
}	

static int remmod_exit(hook_rec_t * hookrec, struct module * mod)
{
	dp_remmod(mod);
	return HOOK_CONTINUE;
}	

int initialise_dprobes_hooks() 
{
	int retval;
	static int done = 0;
	if (done)
		return 0;
	done = 1;

	if ((retval = hook_initialise(SYS_INIT_MODULE)))
		goto err;
	if ((retval = hook_initialise(SYS_INIT_MODULE_ERROR)))
		goto err1;
	if ((retval = hook_initialise(FREE_MODULE)))
		goto err2;
	if ((retval = initialise_asm_dprobes_hooks()))
		goto err3;
	return 0;

err3:	hook_terminate(FREE_MODULE, 0);
err2:	hook_terminate(SYS_INIT_MODULE_ERROR, 0);
err1:	hook_terminate(SYS_INIT_MODULE, 0);
err:	return retval;
}

int terminate_dprobes_hooks()
{
	hook_terminate(SYS_INIT_MODULE, 0);
	hook_terminate(SYS_INIT_MODULE_ERROR, 0);
	hook_terminate(FREE_MODULE, 0);
	terminate_asm_dprobes_hooks();
	return 0;
}

hook_rec_t insmod_hook_rec;
hook_rec_t insmod_err_hook_rec;
hook_rec_t remmod_hook_rec;

int register_dprobes_hooks()
{
	int retval;
	static int done = 0;
	if (done)
		return 0;
	done = 1;
	
	insmod_hook_rec.hook_exit = insmod_exit;
	insmod_hook_rec.hook_exit_name = "dp_insmod_exit";
	if ((retval = hook_exit_register(SYS_INIT_MODULE, &insmod_hook_rec, 0)))
		goto err;

	insmod_err_hook_rec.hook_exit = remmod_exit;
	insmod_hook_rec.hook_exit_name = "dp_remmod_exit";
	if ((retval = hook_exit_register(SYS_INIT_MODULE_ERROR, &insmod_err_hook_rec, 0)))
		goto err1;

	remmod_hook_rec.hook_exit = remmod_exit;
	insmod_hook_rec.hook_exit_name = "dp_remmod_exit";
	if ((retval = hook_exit_register(FREE_MODULE, &remmod_hook_rec, 0)))
		goto err2;

	if ((retval = register_asm_dprobes_hooks()))
		goto err3;

	return 0;
	
err3:	hook_exit_deregister(&remmod_hook_rec);
err2:	hook_exit_deregister(&insmod_err_hook_rec);
err1:	hook_exit_deregister(&insmod_hook_rec);
err:	return retval;
}

int deregister_dprobes_hooks()
{
	hook_exit_deregister(&insmod_hook_rec);
	hook_exit_deregister(&insmod_err_hook_rec);
	hook_exit_deregister(&remmod_hook_rec);
	deregister_asm_dprobes_hooks();
	return 0;
}

int arm_all_dprobes_hooks()
{
	int retval;
	
	if ((retval = hook_exit_arm(&insmod_hook_rec)))
		goto err;
	if ((retval = hook_exit_arm(&insmod_err_hook_rec)))
		goto err1;
	if ((retval = hook_exit_arm(&remmod_hook_rec)))
		goto err2;
	if ((retval = arm_all_asm_dprobes_hooks()))
		goto err3;
	return 0;

err3:	hook_exit_disarm(&remmod_hook_rec);
err2:	hook_exit_disarm(&insmod_err_hook_rec);
err1:	hook_exit_disarm(&insmod_hook_rec);
err:	return retval;
}

int disarm_all_dprobes_hooks()
{
	hook_exit_disarm(&remmod_hook_rec);
	hook_exit_disarm(&insmod_err_hook_rec);
	hook_exit_disarm(&insmod_hook_rec);
	disarm_all_asm_dprobes_hooks();
	return 0;
}

void dprobes_hooks_code_end(void) { }
