/*
 * iobuf.c
 *
 * Keep track of the general-purpose IO-buffer structures used to track
 * abstract kernel-space io buffers.
 * 
 */

#include <linux/iobuf.h>
#include <linux/slab.h>

static kmem_cache_t *kiobuf_cachep;

void end_kio_request(struct kiobuf *kiobuf, int uptodate)
{
	if ((!uptodate) && !kiobuf->errno)
		kiobuf->errno = -EIO;

	if (atomic_dec_and_test(&kiobuf->io_count)) {
		if (kiobuf->end_io)
			kiobuf->end_io(kiobuf);
#if defined(CONFIG_DUMP) || defined(CONFIG_DUMP_MODULE)
			/* the end_io fn should take care of waiters too */
		else
#endif
			wake_up(&kiobuf->wait_queue);
	}
}

static int kiobuf_init(struct kiobuf *iobuf)
{
	int retval;

	init_waitqueue_head(&iobuf->wait_queue);
	iobuf->array_len = 0;
	iobuf->nr_pages = 0;
	iobuf->locked = 0;
	iobuf->dovary = 0;
	atomic_set(&iobuf->io_count, 0);
	iobuf->end_io = NULL;
	iobuf->kio_bh = NULL;
	iobuf->kio_blocks = NULL;
	iobuf->initialized = 0;
	retval = expand_kiobuf(iobuf, KIO_STATIC_PAGES);
	if (retval)
		return retval;
	retval = alloc_kiobuf_bhs(iobuf);
	if (retval) {
		kfree(iobuf->maplist);
		return retval;
	}
	iobuf->initialized = 1;
	return 0;
}

int alloc_kiobuf_bhs(struct kiobuf * kiobuf)
{
	int i;

	kiobuf->kio_blocks = kmalloc(KIO_MAX_SECTORS * sizeof(unsigned long), GFP_KERNEL);
	if (!kiobuf->kio_blocks)
		goto nomem;

	kiobuf->kio_bh = kmalloc(sizeof(struct buffer_head *) * KIO_MAX_SECTORS, GFP_KERNEL);
	if (!kiobuf->kio_bh)
		goto nomem;

	for (i = 0; i < KIO_MAX_SECTORS; i++)
		if (!(kiobuf->kio_bh[i] = kmem_cache_alloc(bh_cachep, SLAB_KERNEL))) {
			while (i--)
				kmem_cache_free(bh_cachep, kiobuf->kio_bh[i]);
			memset(kiobuf->kio_bh, 0, sizeof(struct buffer_head *) * KIO_MAX_SECTORS); /* slow path */
			goto nomem;
		}
	return 0;

nomem:
	free_kiobuf_bhs(kiobuf);
	return -ENOMEM;
}

void free_kiobuf_bhs(struct kiobuf * kiobuf)
{
	int i;

	if (kiobuf->kio_bh) {
		for (i = 0; i < KIO_MAX_SECTORS; i++)
			if (kiobuf->kio_bh[i])
				kmem_cache_free(bh_cachep, kiobuf->kio_bh[i]);
		kfree(kiobuf->kio_bh);
		kiobuf->kio_bh = NULL;
	}
	if (kiobuf->kio_blocks) {
		kfree(kiobuf->kio_blocks);
		kiobuf->kio_blocks = NULL;
	}
}

void kiobuf_ctor(void * objp, kmem_cache_t * cachep, unsigned long flag)
{
	struct kiobuf * iobuf = (struct kiobuf *) objp;
	kiobuf_init(iobuf);
}

void kiobuf_dtor(void * objp, kmem_cache_t * cachep, unsigned long flag)
{
	struct kiobuf * iobuf = (struct kiobuf *) objp;
	if (iobuf->initialized) {
		kfree(iobuf->maplist);
		free_kiobuf_bhs(iobuf);
	}
}

int alloc_kiovec(int nr, struct kiobuf **bufp)
{
	int i;
	struct kiobuf *iobuf;
	
	for (i = 0; i < nr; i++) {
		iobuf = kmem_cache_alloc(kiobuf_cachep, GFP_KERNEL);
		if (!iobuf) {
			free_kiovec(i, bufp);
			return -ENOMEM;
		}
		if (unlikely(!iobuf->initialized)) {
			/* try again to complete previously failed ctor */
			if (unlikely(kiobuf_init(iobuf)))
				goto nomem;
		}	
		bufp[i] = iobuf;
	}
	
	return 0;
nomem:
	kmem_cache_free(kiobuf_cachep, iobuf);
	free_kiovec(i, bufp);
	return -ENOMEM;
}

void free_kiovec(int nr, struct kiobuf **bufp) 
{
	int i;
	struct kiobuf *iobuf;
	
	for (i = 0; i < nr; i++) {
		iobuf = bufp[i];
		init_waitqueue_head(&iobuf->wait_queue);
		iobuf->io_count.counter = 0;
		iobuf->end_io = NULL;
		kmem_cache_free(kiobuf_cachep, iobuf);	
	}
}

int expand_kiobuf(struct kiobuf *iobuf, int wanted)
{
	struct page ** maplist;
	
	if (iobuf->array_len >= wanted)
		return 0;
	
	maplist = (struct page **) kmalloc(wanted * sizeof(struct page **), GFP_KERNEL);
	if (!maplist)
		return -ENOMEM;

	/*
	 * The below check makes sense only under the big kernel lock,
	 * it should be changed in 2.5, current way is bad design.
	 * It's not needed anyways because even 2.4 doesn't hold the
	 * big kernel lock here. Left it here just because it doesn't hurt
	 * either.
	 */
	/* Did it grow while we waited? */
	if (iobuf->array_len >= wanted) {
		kfree(maplist);
		return 0;
	}

	/*
	 * Let's hope every arch implements a memcpy that as expected
	 * first cehcks the size, and then only if the size is > 0 goes to
	 * dereference the buffer pointer.
	 */
	memcpy (maplist, iobuf->maplist, iobuf->array_len * sizeof(struct page **));

	if (iobuf->array_len)
		kfree (iobuf->maplist);
	
	iobuf->maplist   = maplist;
	iobuf->array_len = wanted;
	return 0;
}

void kiobuf_wait_for_io(struct kiobuf *kiobuf)
{
	struct task_struct *tsk = current;
	DECLARE_WAITQUEUE(wait, tsk);

	if (atomic_read(&kiobuf->io_count) == 0)
		return;

	add_wait_queue(&kiobuf->wait_queue, &wait);
repeat:
	set_task_state(tsk, TASK_UNINTERRUPTIBLE);
	if (atomic_read(&kiobuf->io_count) != 0) {
		run_task_queue(&tq_disk);
		schedule();
		if (atomic_read(&kiobuf->io_count) != 0)
			goto repeat;
	}
	tsk->state = TASK_RUNNING;
	remove_wait_queue(&kiobuf->wait_queue, &wait);
}

void __init kio_cache_init(void)
{
	kiobuf_cachep = kmem_cache_create("kiobuf", sizeof(struct kiobuf),
					  0, SLAB_HWCACHE_ALIGN, kiobuf_ctor, kiobuf_dtor);
	if (!kiobuf_cachep)
		panic("Cannot create kiobuf SLAB cache");
}




